//
// Virtual Trackball based on Quaternions
//
// Reto Ltolf,1999,2001.
// E-mail: luetolf@inf.ethz.ch
// URL: http://graphics.ethz.ch/
//


#include "trackball.h"


//globals
static GLint     tb_timestamp;    //variable to hold time stamp
static GLuint    tb_width;        //stores viewer window width
static GLuint    tb_height;       //stores viewer window height
static GLuint    tb_button  = GLUT_MIDDLE_BUTTON;  //default mouse button for rotation
static GLboolean tb_animate = GL_FALSE;  //flag indicating whether animation is enabled
static GLboolean tb_rotate  = GL_FALSE;  //flag denoting ongoing rotation
static GLfloat   tb_matrix[4][4]; //current rotation matrix
static GLfloat   tb_lastquat[4];  //last quaternion
static GLfloat   tb_newquat[4];   //new quaternion
static GLfloat   tb_lastpos[3];   //last position on virtual trackball (3D vector)


//internal vector functions
void _TB_VEC_copy(const GLfloat *v1, GLfloat *v2)
//copies vector data from v1 to v2
{
  register int i;

  for (i=0;i<3;i++) v2[i] = v1[i];
}

void _TB_VEC_scale(GLfloat *v, GLfloat s)
//scales vector v with scalar value s
{
  v[0] *= s;
  v[1] *= s;
  v[2] *= s;
}

void _TB_VEC_add(const GLfloat *v1, const GLfloat *v2, GLfloat *s)
//adds v1 to v2 and returns result in s
{
  s[0] = v1[0] + v2[0];
  s[1] = v1[1] + v2[1];
  s[2] = v1[2] + v2[2];
}

GLfloat _TB_VEC_dot(const GLfloat *v1, const GLfloat *v2)
//returns dot product of vectors v1 and v2
{
  return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
}

void _TB_VEC_cross(const GLfloat *v1, const GLfloat *v2, GLfloat *cross)
//calculates cross product of vectors v1 and v2 and returns
//result in parameter cross
{
  GLfloat t[3];

  t[0] = (v1[1]*v2[2]) - (v1[2]*v2[1]);
  t[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);
  t[2] = (v1[0]*v2[1]) - (v1[1]*v2[0]);
  _TB_VEC_copy(t, cross);
}

void _TB_VEC_project(GLfloat v[3], int x, int y)
//projects the mouse position x,y to virtual trackball
//and returns the 3D position in vector v
{
  GLfloat x_coord;
  GLfloat y_coord;

  //normalize both x,y to be in [-1.0,1.0]
  //!! the upper left corner of a window handled
  //   by GLUT is (0,0) - this is not consistent
  //   with OpenGL... ;-(
  //   (read section "1.5 Conventions" on GLUT API web page)
  /// Add your code here
  

  //make sure x_coord and y_coord are within
  //[-1.0,1.0] after the above calculations
  /// Add your code here


  //calculate z coordinate using TB_TRACKBALL_RADIUS defined
  //in trackball.h and set components of v to be the calculated
  //values
  /// Add your code here

}


//internal quaternion functions
void _TB_QUAT_norm(GLfloat quat[4])
//normalizes the passed quaternion quat
{
  register int i;
  GLfloat magnitude;

  //a unit quaternion is normalized just
  //like a unit vector...
  /// Add your code here

}

void _TB_QUAT_mult(GLfloat quat1[4], GLfloat quat2[4], GLfloat dest[4])
//multiplicates quaternions quat1 and quat2 and copies
//result to dest
//!! make sure a caller can use the same parameters
//   for quat2 and dest ...
//   (e.g. see invocation of this function in TB_motion)
{
  GLfloat t1[4], t2[4], t3[4], tf[4];

  //multiply the things using the given formula
  //and using the given vector functions in this module
  /// Add your code here
 

  //copy result to dest
  /// Add your code here


  //normalize calculated quaternion
  _TB_QUAT_norm(dest);
}

void _TB_QUAT_buildQuat(GLfloat quat[4], GLfloat v[3])
//build new quaternion with given new position (parameter v)
//and the last position stored in 'tb_lastpos'
//return calculated quaternion in parameter quat
{
  GLfloat a[3];
  GLfloat dist[3];
  GLfloat phi;

  //return [0,0,0,1]-quaternion if mouse did not move
  //(i.e. new and last positions are identical)
  if ( v[0]/* == ... */ ) {
    /// Add your code here
 
    return;
  }

  //calculate distance between new and last
  //3D position on virtual trackball
  /// Add your code here


  //calculate angle phi with the above calculated
  //distance (using again TB_TRACKBALL_RADIUS)
  //!! make sure the value passed to function 'asin'
  //   lives within [-1.0,1.0] ...
  /// Add your code here
 

  //calculate rotation axis
  /// Add your code here

  //normalize rotation axis
  /// Add your code here

  //build quaternion in parameter quat
  /// Add your code here

}


void _TB_QUAT_buildMatrix(GLfloat quat[4])
//builds rotation matrix with given quaternion
{
  tb_matrix[0][0] = 1.0 - 2.0 * (quat[1]*quat[1] + quat[2]*quat[2]);
  tb_matrix[1][0] =       2.0 * (quat[0]*quat[1] - quat[2]*quat[3]);
  tb_matrix[2][0] =       2.0 * (quat[2]*quat[0] + quat[1]*quat[3]);
  tb_matrix[3][0] = 0.0;

  tb_matrix[0][1] =       2.0 * (quat[0]*quat[1] + quat[2]*quat[3]);
  tb_matrix[1][1] = 1.0 - 2.0 * (quat[2]*quat[2] + quat[0]*quat[0]);
  tb_matrix[2][1] =       2.0 * (quat[1]*quat[2] - quat[0]*quat[3]);
  tb_matrix[3][1] = 0.0;

  tb_matrix[0][2] =       2.0 * (quat[2]*quat[0] - quat[1]*quat[3]);
  tb_matrix[1][2] =       2.0 * (quat[1]*quat[2] + quat[0]*quat[3]);
  tb_matrix[2][2] = 1.0 - 2.0 * (quat[1]*quat[1] + quat[0]*quat[0]);
  tb_matrix[3][2] = 0.0;

  tb_matrix[0][3] = 0.0;
  tb_matrix[1][3] = 0.0;
  tb_matrix[2][3] = 0.0;
  tb_matrix[3][3] = 1.0;
}


//internal functions
static void _TB_idle_callback(void)
//idle callback to be registered when doing
//animation
{
  //multiply new quaternion forever ... ;-)
  _TB_QUAT_mult(tb_newquat, tb_lastquat, tb_lastquat);
  glutPostRedisplay();
}

void _TB_start(int x, int y, GLint time)
//initialize and start rotation
//(see TB_mouse)
{
  tb_timestamp = time;
  tb_rotate = GL_TRUE;

  _TB_VEC_project(tb_lastpos, x, y);
}

void _TB_stop(GLint time)
//stoppes rotation and registeres or de-registeres
//idle callback
//(see TB_mouse)
{
  tb_rotate = GL_FALSE;
  if (time==tb_timestamp && tb_animate) glutIdleFunc(_TB_idle_callback);
  else if (tb_animate) glutIdleFunc(NULL);
}


//public functions
void TB_init(GLuint button)
//initializes trackball module
{
  GLint current_mode;

  tb_button = button;

  tb_lastquat[0] = 0.0;
  tb_lastquat[1] = 0.0;
  tb_lastquat[2] = 0.0;
  tb_lastquat[3] = 1.0;    

  //an easy way to load identity matrix
  //to tb_matrix ... ;-)
  glGetIntegerv(GL_MATRIX_MODE, &current_mode);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glGetFloatv(GL_MODELVIEW_MATRIX, &tb_matrix[0][0]);
  glPopMatrix();
  glMatrixMode(current_mode);
}

void TB_animate(GLboolean animate)
//sets animation flag
{
  tb_animate = animate;
}

void TB_mouse(int button, int state, int x, int y)
//reacts upon mouse stats
//(to be called from mouse callback function)
{
  if (button==tb_button && state==GLUT_DOWN) _TB_start(x, y, glutGet(GLUT_ELAPSED_TIME));
  else if (button==tb_button && state==GLUT_UP) _TB_stop(glutGet(GLUT_ELAPSED_TIME));
}

void TB_motion(int x, int y)
//reacts upon motion stats
//(to be called from motion callback function)
{
  GLfloat new_pos[3];

  //return immediately if rotation is not enabled
  if (!tb_rotate) return;
  
  //project current mouse position to virtual trackball
  _TB_VEC_project(new_pos, x, y);
  //build new quaternion
  _TB_QUAT_buildQuat(tb_newquat, new_pos);
  //save new position in 'tb_last'
  _TB_VEC_copy(new_pos, tb_lastpos);
  //multiply new quaternion to last one
  _TB_QUAT_mult(tb_newquat, tb_lastquat, tb_lastquat);

  //get time stamp
  tb_timestamp = glutGet(GLUT_ELAPSED_TIME);
  glutPostRedisplay();
}

void TB_reshape(int w, int h)
//stores new window size in internal variables
//(to be called from reshape callback function)
{
  tb_width = w;
  tb_height = h;
}

void TB_matrix(void)
//writes current rotation matrix to modelview stack
//(to be called from display callback function)
{
  GLint current_mode;

  //build current rotation matrix
  _TB_QUAT_buildMatrix(tb_lastquat);

  //multiply current rotation matrix to modelview stack
  glGetIntegerv(GL_MATRIX_MODE, &current_mode);
  glMatrixMode(GL_MODELVIEW);
  glMultMatrixf(&tb_matrix[0][0]);
  glMatrixMode(current_mode);
}
